/*---------------------------------------------------------------------
  CableTester.c by Jack Botner
  
  To test various cable configurations, up to 8 conductors.
  - 2x16 LCD display
  - Rotary encoder with push button

  Processor: PIC16F877A
  Clock: 3.57 MHz crystal
  
  This code is provided for EDUCATIONAL purposes ONLY. There is no
  warranty and no support.
---------------------------------------------------------------------*/
#include <system.h>
#include <memory.h>
#include "CableTester.h"

#pragma CLOCK_FREQ      3579545
#pragma DATA _CONFIG, _CP_OFF & _HS_OSC & _WRT_OFF & _CPD_OFF & _LVP_OFF & _BODEN_OFF & _PWRTE_ON & _WDT_OFF

// LCD stuff: for LCD on PORTD RD0-RD3 (data), RD4 (RW), RD5 (RS), RD6 (EN)

#define LCD_ARGS  1,  /* Interface type: mode 0 = 8bit, 1 = 4bit(low nibble), 2 = 4bit(upper nibble) */ \
      0,              /* Use busy signal: 1 = use busy, 0 = use time delays */\
      PORTD, TRISD,   /* Data port and data port tris register */ \
      PORTD, TRISD,   /* Control port and control port tris register */ \
      5,              /* Bit number of control port is connected to RS */ \
      4,              /* Bit number of control port is connected to RW */ \
      6               /* Bit number of control port is connected to Enable */

#include <lcd_driver.h> // include the LCD template code

// Global Variables
volatile unsigned char ucEventFlags=0;		// event flags
volatile unsigned char ucDebouncedState=0;	// debounced state of the switches
volatile unsigned char ucPreviousState=0;	// previous switch state

static volatile unsigned char ucConductors=0;		// 1..8
static volatile unsigned char ucMap[8];

// Internal functions
static void InitCPU(void);
static void report1( unsigned char ucIndex );
static unsigned char report2(void);
static void test_cable(void);
static unsigned char testinput( unsigned char ucIndex );

/*----------------------------------------------------------------
  main()
----------------------------------------------------------------*/
void main()
{
  rom char  *cIntroMsg = "Cable Tester 1.0";
  rom char  *pcPrompt1 = "Conductors?";
  rom char  *pcPrompt2 = "Testing Cable...";
  unsigned char ucRun, ucIndex, ucRc;

  InitCPU();                // initialize hardware and start timer1
  lcd_setup();              // initialize LCD
  lcd_clear();

  lprintf( cIntroMsg );     // display intro message
  delay_s( 2 );             // wait a bit

  while (1)
  {
    // Step 1: get the number of conductors
    
    lcd_clear();
    lprintf( pcPrompt1 );     // display prompt
    lcd_gotoxy( 0, 1 );
    lprintf( "%u", (unsigned int) ucConductors );
    ucRun = 1;
    
    while ( ucRun )
    {
	  if ( test_bit( ucEventFlags, EVENT_SWITCH_CHANGE ) )
      {
        analyze_switch();       // convert switch change to events
        clear_bit( ucEventFlags, EVENT_SWITCH_CHANGE );
      }
      
      if ( test_bit( ucEventFlags, EVENT_ENCODER_UP ) )
      {
        if ( ucConductors < 8 )
        {
          ++ucConductors;
          lcd_gotoxy( 0, 1 );
          lprintf( "%u", (unsigned int) ucConductors );
        }
        clear_bit( ucEventFlags, EVENT_ENCODER_UP );
      }

      if ( test_bit( ucEventFlags, EVENT_ENCODER_DOWN ) )
      {
        if ( ucConductors > 1 )
        {
          --ucConductors;
          lcd_gotoxy( 0, 1 );
          lprintf( "%u", (unsigned int) ucConductors );
        }
        clear_bit( ucEventFlags, EVENT_ENCODER_DOWN );
      }  

      if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
      { 
		if ( ucConductors )
          ucRun = 0;
        clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
      }  
    }
    
    // Step 2: test the cable
    
    lcd_clear();
    lprintf( pcPrompt2 );     // display prompt
    test_cable();

    // Step 3: report the results

	report1( 0 );		// display the initial report
    ucIndex = 0;
    ucRun = 1;
    while ( ucRun )
    {
	  if ( test_bit( ucEventFlags, EVENT_SWITCH_CHANGE ) )
      {
        analyze_switch();         // convert switch change to events
        clear_bit( ucEventFlags, EVENT_SWITCH_CHANGE );
      }
      
      if ( test_bit( ucEventFlags, EVENT_ENCODER_UP ) )
      {
        if ( ( ucIndex + 1 ) < ( ucConductors - 1 ) )
        {
          ++ucIndex;
          report1( ucIndex );
        }
        clear_bit( ucEventFlags, EVENT_ENCODER_UP );
      }

      if ( test_bit( ucEventFlags, EVENT_ENCODER_DOWN ) )
      {
        if ( ucIndex )
        {
          --ucIndex;
          report1( ucIndex );
        }
        clear_bit( ucEventFlags, EVENT_ENCODER_DOWN );
      }  

      if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
      { 
        ucRun = 0;		// quit
        clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
      }  
    }

    // Step 4: any special report

	ucRc = report2();
	if ( ucRc )			// there was a report
	{  
	  // Wait for the button press
      while (1)
      {
	    if ( test_bit( ucEventFlags, EVENT_SWITCH_CHANGE ) )
        {
          analyze_switch();         // convert switch change to events
          clear_bit( ucEventFlags, EVENT_SWITCH_CHANGE );
        }

        if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
        { 
          clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
          break;
        }      
	  }
	}  
  }
}

/*---------------------------------------------------------------------
  InitCPU()
---------------------------------------------------------------------*/
static void InitCPU()
{
  // PORTA connections as follows:
  //  RA0 - (input) Cable1
  //  RA1 - (input) Cable2
  //  RA2 - (input) Cable3
  //  RA3 - (input) Cable4
  //  RA4 - (output) not used
  //  RA5 - (input) Cable5

  porta = 0;
  trisa = 00101111b;            // 1=input, 0=output

  // PORTB connections as follows:
  //  RB0 - (input) Rotary Encoder OUT A *
  //  RB1 - (input) Rotary Encoder OUT B *
  //  RB2 - (input) Rotary Encoder pushbutton *
  //  RB3 - (output) not used
  //  RB4 - (output) not used
  //  RB5 - (output) not used
  //  RB6 - (input) PGC
  //  RB7 - (input) PGD
  // * = pull-up required
  
  portb = 0;
  trisb = 11000111b;				// 1=input, 0=output
  clear_bit( option_reg, NOT_RBPU ); // enable weak pull-ups on input pins

  // PORTC connections as follows:
  //  RC0..RC7 - (output) Cable1..8

  portc = 0;
  trisc = 00000000b;            // 1=input, 0=output

  // PORTD connections as follows:
  //  RD0 - LCD DB4 (output)
  //  RD1 - LCD DB5 (output)
  //  RD2 - LCD DB6 (output)
  //  RD3 - LCD DB7 (output)
  //  RD4 - LCD RW (output)
  //  RD5 - LCD RS (output)
  //  RD6 - LCD EN (output)
  //  RD7 - (output) diagnostic LED

  portd = 0;
  trisd = 00000000b;            // all bits output (0)

  // PORTE connections as follows:
  //  RE0 - (input) Cable6
  //  RE1 - (input) Cable7
  //  RE2 - (input) Cable8

  porte = 0;
  trise = 00000111b;            // 1=input, 0=output

  adcon0 = 0;					// A/D off
  adcon1 = 00000110b;			// all ports digital (!)

  // Timer1 initialization

  tmr1h = TIMER1_H;
  tmr1l = TIMER1_L;
  t1con = 00000001b;            // 1:1 prescaleR; int. clock; timer enabled

  // Enable interrupts for Timer1

  set_bit( intcon, GIE );       // enable global interrupts
  set_bit( intcon, PEIE );      // enable peripheral interrupts
  set_bit( pie1, TMR1IE );      // Timer 1 interrupts
}

/*---------------------------------------------------------------------
  interrupt()
  
  Timer1 is used to generate 4 ms interrupts for switch debouncing.

  Timer1 counts down 16 bits, giving a maximum of 65536 counts. Counting
  is done by the processor, each instruction cycle. Since the clock is
  3.58 mHz, the timer increments every 1.11 microsecond. With 1:1
  prescaler we need to count 3600 to get 4 ms.
---------------------------------------------------------------------*/
void interrupt( void )
{
  static volatile unsigned char ucCounter=0;
   
  if ( test_bit( pir1, TMR1IF ) )
  {
    // Process switch input 
    debounce_switch();			// sets ucDebouncedState
    if ( ucDebouncedState != ucPreviousState )
	  set_bit( ucEventFlags, EVENT_SWITCH_CHANGE );

    // Manage the diagnostic led
    ++ucCounter;
    if ( ucCounter == 125 )		// 500 ms
    {
      if ( test_bit( portd, 7 ) )
        clear_bit( portd, 7 );
      else  
        set_bit( portd, 7 );
        
      ucCounter = 0;
    }
    
    // Reset the timer
    tmr1h = TIMER1_H;
    tmr1l = TIMER1_L;
    clear_bit( pir1, TMR1IF );    // reset the TMR1 overflow interrupt flag
  }
}

/*---------------------------------------------------------------------
  test_cable()
  
  Tests the cable and creates a report in ucMap table. Each byte
  is the report for each conductor. Each bit in the byte is a map
  to hits at the other end of the cable.
  
  Input: ucConductors (number of conductors, 1..8)
  InOut: ucMap[]
---------------------------------------------------------------------*/
static void test_cable()
{
  unsigned char	ii, jj;

  for ( ii=0; ii<ucConductors; ++ii )
  {
    set_bit( portc, ii );		// raise port
    ucMap[ii] = 0;
    for ( jj=0; jj<ucConductors; ++jj )
    {
      if ( testinput( jj ) )
      {
        set_bit( ucMap[ii], jj );
      }  
    }  
    clear_bit( portc, ii );	  // drop port
  }      
}

/*---------------------------------------------------------------------
  testinput(...)
  
  Returns: 0 or 1 depending on the state of the input pin
		   defined by ucIndex.
---------------------------------------------------------------------*/
static unsigned char testinput( unsigned char ucIndex )
{
  unsigned char ucRc=0;
  
  switch ( ucIndex )
  {
    case 0:
      ucRc = test_bit( porta, 0 );
      break;
    case 1:
      ucRc = test_bit( porta, 1 );
      break;
    case 2:
      ucRc = test_bit( porta, 2 );
      break;
    case 3:
      ucRc = test_bit( porta, 3 );
      break;
    case 4:
      ucRc = test_bit( porta, 5 );
      break;
    case 5:
      ucRc = test_bit( porte, 0 );
      break;
    case 6:
      ucRc = test_bit( porte, 1 );
      break;
    case 7:
      ucRc = test_bit( porte, 2 );
      break;
  }
  
  return ucRc;
}

/*---------------------------------------------------------------------
  report1(...)
 
  Reports one or two lines of information of the cable mapping.
  The report of the conductor specified by ucIndex is placed on the
  1st line of the lcd. The report of the next conductor, if it
  exists, is placed on the 2nd line of the lcd.
 
  Input: ucIndex (0..6)
		 ucConductors (1..8)
		 ucMap[8]
---------------------------------------------------------------------*/
static void report1( unsigned char ucIndex )
{
  unsigned char ii;
  
  lcd_clear();

  // Avoid processing invalid data
  if ( !ucConductors ) return;
  if ( ucIndex )
    if ( ucIndex >= ( ucConductors - 1 ) ) return;
  if ( ucIndex > 6 ) return;

  // Report the 1st line
  
  lprintf( "C%u: ", (unsigned int) ( ucIndex + 1 ) );
  if ( ucMap[ucIndex] )
  {
    for ( ii=0; ii<ucConductors; ++ii )
    {
      if ( test_bit( ucMap[ucIndex], ii ) )
        lprintf( "%u ", (unsigned int) ( ii + 1 ) );
    }
  }
  else  
    lprintf( "no mapping" );

  // Report the 2nd line

  if ( ucConductors < 2 ) return;	// done
  ++ucIndex;
  lcd_gotoxy( 0, 1 );  
  lprintf( "C%u: ", (unsigned int) ( ucIndex + 1 ) );

  if ( ucMap[ucIndex] )
  {
    for ( ii=0; ii<ucConductors; ++ii )
    {
      if ( test_bit( ucMap[ucIndex], ii ) )
        lprintf( "%u ", (unsigned int) ( ii + 1 ) );
    }
  }
  else  
    lprintf( "no mapping" );
}

/*---------------------------------------------------------------------
  report2()
 
  This module analyzes the data for a special report, such as
  a network cable.
  
  Input: ucConductors (1..8)
		 ucMap[8]
		 
  Returns: 0 = no report, 1 = report displayed		 
---------------------------------------------------------------------*/
static unsigned char report2()
{
  rom char  *pcMsg1 = "Twisted Pair";
  rom char  *pcMsg2 = "Patch Cable";
  rom char  *pcMsg3 = "Cross Cable";
  rom char  *pcMsg4 = "100BaseT4";

  static const unsigned char ucPatch[8] = 
  { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };
  static const unsigned char ucCross[8] = 
  { 0x04, 0x20, 0x01, 0x08, 0x10, 0x02, 0x40, 0x80 };
  static const unsigned char uc100BT4[8] = 
  { 0x04, 0x20, 0x01, 0x40, 0x80, 0x02, 0x08, 0x10 };

  if ( ucConductors != 8 ) return 0;	// not network cable
  if ( !memcmp( ucMap, ucPatch, 8 ) )
  {
    lcd_clear();
    lprintf( pcMsg1 );
    lcd_gotoxy( 0, 1 );  
    lprintf( pcMsg2 );
    return 1;
  }
 
  if ( !memcmp( ucMap, ucCross, 8 ) )
  {
    lcd_clear();
    lprintf( pcMsg1 );
    lcd_gotoxy( 0, 1 );  
    lprintf( pcMsg3 );
    return 1;
  }
 
  if ( !memcmp( ucMap, uc100BT4, 8 ) )
  {
    lcd_clear();
    lprintf( pcMsg1 );
    lcd_gotoxy( 0, 1 );  
    lprintf( pcMsg4 );
    return 1;
  }
  
  return 0;  
}
